#!/usr/local/bin/perl
#
# name  : server.pl
# author: jolly - who else
#
#
#	This is the main part of the distributed system - it is well documented, so 
#	that you can apply changes where needed.
#	Be aware of the global vars ! - they gave me a hard time - one of the few disadvanteges
#	using perl. But for testpurposes you can use the perl -w option
#

# #######
# configurable variables
# #######

$debug              =2;     # debug level (0,1,2,3)
$pingintervall      =1;     # ping intervall in seconds
$routeintervall     =10;    # rerouting intervall in pingintervalls
$rerouteintervall   =120;   # rerouting intervall in pingintervalls when the distant routers
                            # are asked whom they know - in case we missed some important 
                            # messages ( wich should not happen ) 
$reliable           =85;    # percent of messagereliability needed for routing


require "netstart.pl";      # create network socket for receiving messages and setup
                            # remote servers ( reading the configuration files )
require "routines.pl";      # subroutines for messag sending/receiving ...
require "../example/handler.pl";		# This file includes all the mechanisms a specific implementation
                            			# has to do ( like storing, deleting ... whatever )

# The following variables are now configured
#
# $myhostname               ascii name
# $myhostaddr               integer value stored as ascii ( for use in recv(2) )
# $myboottime               a variable $fromid{$from} wich is used to consider 
#                           if a remote host has been rebooted during runtime 
#                           ( this var is our bootime wich is send)
#
# $hardrouted{addr}     =paddr      ascci values of all hard routed servers ( are
# $hostname{hardrouted}             read from the configuration file ) 


# Other global vars
#
# $from,$fromid,$origin,$origid,$order,$message
# 
# $broadcast{$addr}     =&encodehosts(@hosts);      a ascii string wich has to be decoded via
#                                                   (@hosts)=&decodehosts($broadcast{$addr})
#                                                   contains all hosts wich have to be messaged
#                                                   when a message from $addr arrives.
#
# $router{fasel}        =hardrouted;                holds the hardrouter we rely upon when 
#                                                   awaing a message from host `fasel'
#
# $hops{addr}                                       number of servers between us and host
#                                                   addr ( direct hosts == 1 hop )
#
# $speed{addr}                                      theoretical speed till messages arrive 
#                                                   from host addr.
#
# $rely{addr}                                       reliablility that messages to host
#                                                   'addr' will arrive.

# Variables where the pending time of the pingintervalls is stored in case
# we get messages between pings
# 
# $selecttime       # time to select(2)
# $selectend  
# $intervallcounter # number of pingintervalls raised


# reliable messages are stored in the following variables. If we don't get any
# answer regarding the broadcasting of those messages they will be resent 
#
# $raddr{origid}
# $rorigin{origid}
# $rorder{origid}
# $rmessage{rorigid}




# ########
# send every pingintervall a ping, recalculate routings every $routeintervall
# and receive messages inbetween
# ########

&HANDLER_Init;  # give our handler a chance to update the directories ... 

$selectend=time()+$pingintervall;
while(1)
{
    $selecttime=$selectend-time();
    $pending=select($riin=$rin,undef,undef,$selecttime) if ($selecttime>0) || ($pending=0);
    
    if($pending) 
    {
        &handlemessage;
    }
    else
    {
        $selectend=time()+$pingintervall;
        $intervallcounter=($intervallcounter+1)%$routeintervall;    
        $intervall2counter=($intervall2counter+1)%$rerouteintervall;    
        &routeintervall if !$intervallcounter;
        &rerouteintervall if !$intervall2counter;
        &pingintervall;
        &HANDLER_Heartbeat;
    }
}

# ########
# end of main program 
# ########





sub handlemessage
{
    ($from,$fromid,$origin,$origid,$order,$message)=&recvmessage;
    $_=$order;


    # ########
    # if the remoteserver has been rebooted it is no longer reliable, reliable messages
    # have to be removed and we have to inform the other one that we don't trust him 
    # anymore
    # ########
    if( $fromid{$from} != $fromid )
    {
        print STDERR "Server $hostname{$from} has been rebootet\n" if $debug;
        &removehost($from);
        $fromid{$from}=$fromid;

        return undef;
    }



    # ########
    # when a reliable message returns we have to remove that message from our pending-list.
    # ########
    if( $raddr{$origid} && ($rorigin{$origid} eq $origin) && ($raddr{$origid} eq $from) )
    {
        $raddr{$origid}=undef;
        $rorigin{$origid}=undef;
        $rorder{$origid}=undef;
        $rmessage{$rorigid}=undef;

        return undef if( /^got message$/ );
    }



    # ########
    # when we don't trust a server any longer ( rely{$from} droped below reliable )
    # that server is informed about that by getting a 'remove routing ping' instead of
    # normal 'ping' then he sends a 'remove routing pong' so that we know, he knows that
    # he's not longer reliable for us.
    # ########
    if(/^remove routing ping$/)
    {
        &removehost($from);
        &umessage($from,$origin,$origid,"remove routing pong",$message);
        return undef;
    }
    if(/^remove routing pong$/)
    {
        $removerouting{$from}=undef;
        return undef;
    }



    # ########
    # a 'ping' will be answerd by a pong , a pong will be attached to a list containing 
    # the time the ping - pong timedifference
    # ########
    if(/^ping$/)
    {
        &umessage($from,$origin,$origid,"pong",undef);
        return undef;
    }
    if(/^pong$/)
    {
        $pongrecv{$from}.=pack("i i",$origid,time());
        return undef;
    }



    # ########
    # when a former reliable server is now unreliable but did not answer a 
    # 'remove routing pong' we just drop that messages
    # ########
    if($rely{$from} <= $reliable)
    {
        print STDERR "An unreliable server $hostname{$from} messaged.\n" if $debug;
        return undef;
    }



# ########
# the routing begins here - now are the routing messages are handled.
# ########



    # ########
    # 'whom do you know' is a question which we will answer in sending all known 
    # servers to the sender excluding the ones which are routed over that server.
    # ########

    if(/^whom do you know$/)
    {
        undef @hosts;
        undef $message;

        foreach $mom (keys(%router))
        {
            next if $mom eq $myhostaddr;
            if( ($router{$mom}) && ($router{$mom} ne $from) && ($rely{$router{$mom}}>$reliable) )
            {
                $message.=pack("a4 s f",$mom,$hops{$mom},$speed{$mom});
            }
        }
        &umessage($from,$origin,$origid,"i know",$message);
        
        return undef;
    }



    # ########
    # 'i know' is the answer on 'whom do you know' so that we will be able to distinguish 
    # if the server (from) known anything we don't know. In that case we want to be 
    # informed if those - for us at the moment unknown - servers send messages
    # ########

    if(/^i know$/)
    {   
        undef @unknown;
        undef $newroute;
        
        while(length($message)) 
        {
            ($addr,$hops,$speed)=unpack("a4 s f",$message);
            next if $addr eq $myhostaddr;
            $speed+=$speed{from};
            $hops++;
            if( (!$router{$addr}) || ($hops{$addr}>$hops) )
            {
                ($name)=gethostbyaddr($addr,&PF_INET);
                $hostname{$addr}=$name if !$hostname{$addr};
                $newroute.=substr($message,0,6);
            }
            $message=substr($message,10);
        }
        &rmessage($from,$myhostaddr,++$myrmessageid,"add broadcast",$newroute) if $newroute;
        &umessage($from,$origin,$origid,"got message",undef) if $origin ne $myhostaddr;

        return undef;
    }



    # ########
    # 'add broadcast' is a message we get from a server who doesn't has access to 
    # specific hosts except we'll message that host in case we get a message from these
    # for the senders unknown hosts
    # ########
    if(/^add broadcast$/)
    {
        undef $newroute;

        while($message)
        {
            ($addr,$hops)=unpack("a4 s",$message);
            if( ($router{$addr}) && ($router{$addr} ne $from) && ($hops{$addr}==$hops) )
            {
                undef @hosts;

                @mom=&decodehosts($broadcast{$addr});
                @hosts=grep(! /$from/,@mom) if @mom;
                push(@hosts,$from);
                $broadcast{$addr}=&encodehosts(@hosts);
                $newroute.=substr($message,0,6);
                $newroute.=pack("f",$speed{$addr});
            }
            $message=substr($message,6);
        }
        &umessage($from,$origin,$origid,"added broadcasts",$newroute);

        return undef;
    }



    # ########
    # reply on a 'add broadcast' is a 'added broadcasts'. Know we know for sure that
    # the sender will forward every message from the hosts listed.
    # ########

    if(/^added broadcasts$/)
    {
        &addroutes($from,$message);

        return undef;
    }



    # ########
    # this means that in case that our databases got messed we can resyncronizise.
    # ########
    if(/^everything i route for you$/)
    {
        undef %oldroute;

        foreach $mom (keys(%router))
        {
            if( ($mom eq $from) && ($router{$mom} ne $from) )
            {
                $router{$mom}=undef;
                $hops{$mom}=undef;
            }
        }
        &addroutes($from,$message);

        return undef;
    }



    # ########
    # delete the listed hosts from the messaging list. Whenever in the future a listed
    # host will send us a message the sender won't be informed.
    # ########
    if(/^delete broadcast$/)
    {
        @hosts=&decodehosts($message);
        foreach $mom (@hosts)
        {
            $broadcast{$mom}=&encodehosts( grep(! /$from/,&decodehosts($broadcast{$mom})) );
        }
        &umessage($from,$origin,$origid,"broadcast deleted",undef);

        return undef;
    }



    # ########
    # a list of hosts is not longer known to the sender - but we relied on that 
    # server. So we will all other reliable servers if they could do us the
    # favour and sending us their latest routing tables.
    # ########
    if(/^delete routes$/)
    {
        @hosts=&decodehosts($message);
        foreach $mom (@hosts)
        {
            $router{$mom}=undef if $router{$mom} eq $from;
        }
        &umessage($from,$origin,$origid,"got message",undef);
        &broadcastintegrety($from);     

        foreach $mom (keys(%hardrouted))
        {
            if( ($mom ne $from) && ($rely{$mom}>=$reliable))
            {
                &rmessage($mom,$myhostaddr,++$myrmessageid,"whom do you know",undef);
            }
        }

        return undef;
    }



    # ########
    # This is now the acutal broadcast wich will be forwarded to every server.
    # attention this message is only runtime reliable. For nonruntime reliable
    # messages the messages have to be stored and the servers have to be resynced
    # everytime the server changes from unreliable to reliable.
    # ########
    if(/^broadcast$/ )
    {
        &umessage($from,$origin,$origid,"got message",undef);
        return undef if &isreceived($origin,$origid);
        &addreceived($origin,$origid);
        
        @hosts=&decodehosts($broadcast{$origin}); 
        foreach $mom (@hosts)
        {
            &rmessage($mom,$origin,$origid,"broadcast",$message);
        }
        
        # ########
        # here the subroutine, handling the message get called
        # ########
        &HANDLER_GotBroadcast($from,$origin,$origid,$message);
        
        return undef;
    }


    
    # ########
    # command will now be handled
    # ########
    if(/^command:(.*)$/)
    {
        &umessage($from,$origin,$origid,"got message",undef);
        return undef if &isreceived($origin,$origid); 
        &addreceived($origin,$origid);
        
        &HANDLER_GotCommand($from,$origin,$origid,$1,$message);

        return undef;
    }

    return undef;
}






# ########
# this subroutine will be called every pingintervall of time. At the moment 
# this procedure does only a real ping / 'remove routing' ping.
# ########

sub pingintervall
{
    foreach $whom (keys(%hardrouted))
    {
        $mymessageid++;
        if($removerouting{$whom})
        {
            &umessage($whom,$myhostaddr,$mymessageid,"remove routing ping",$removerouting{$whom});
        }
        else
        {
            &umessage($whom,$myhostaddr,$mymessageid,"ping",undef);
            $pingsend{$whom}.=pack("i i",$mymessageid,time());
        }
    }

    return undef;
}   


sub rerouteintervall
{
    foreach $whom ( keys(%hardrouted) )
    {
        if($rely{$whom}>=$reliable)
        {
            $mymessageid++;
            &umessage($whom,$myhostaddr,$mymessageid,"whom do you know",undef);
        }
    }
}


sub routeintervall
{
    $accept=time()-($routeintervall*$pingintervall);

    # ########
    # reliable messages that were not received should be send again
    # this will be done in the for loop.
    # ########
    foreach $mom (keys(%rmessage))
    {
        if($mom && $router{ $raddr{$mom} } )
        {
            &umessage($raddr{$mom},$rorigin{$mom},$mom,$rorder{$mom},$rmessage{$mom}) ;
        }
        else
        {   
            $raddr{$mom}=undef;
            $rorigin{$mom}=undef;
            $rorder{$mom}=undef;
            $rmessage{$mom}=undef;
        }
    }


    # ########
    # calculate connection reliablility this is very complicated - but is eaier 
    # than the code looks. 
    # The foreach recalculates the routing reliability for every hardrouted host
    # ########
    foreach $from (keys(%hardrouted))
    {
        undef @send,@recv;



        # ########
        # calculate the avarage time a ping was answered by a pong;
        # how many were lost and remove pings that are toooooo old, cause we
        # are not interested how good the line was 5 hours ago.
        # ########
        $len=length($pingsend{$from})/4;
        @send=unpack("i$len",$pingsend{$from}) if $len;
        $pingsend{$from}=undef;
        undef %msgs;

        while( @send )
        {
            $t=pop(@send);
            $m=pop(@send);
            if($t>$accept)
            {
                $msgs{$m}=$t;
                $pingsend{$from}.=pack("i i",$m,$t);
            }
        }
    

        $len=length($pongrecv{$from})/4;
        @recv=unpack("i$len",$pongrecv{$from}) if $len;
        $pongrecv{$from}=undef;
        undef %msgr;

        while( @recv )
        {
            $t=pop(@recv);
            $m=pop(@recv);
            if($t>$accept)
            {
                $msgr{$m}=$t;
                $pongrecv{$from}.=pack("i i",$m,$t);
            }
        }
    
        
        # ########
        # let's calculate who many packets were lost and 
        # what the usual resending time was.
        # ########

        undef   $count;
        undef   $median;
        undef   $lost;
        foreach $mom ( keys(%msgs) )
        {
            if( $msgr{$mom} )
            {
                $median+=$msgr{$mom}-$msgs{$mom};
                $count++;
            }
            else
            {
                $lost++;
            }
        }
        if($count)
        {
            $rely=100*$count/($count+$lost);
            $speed=$median/$count;
        }
        else
        {
            $rely=0;
            $speed=0;
        }   


        
        # ########
        # we only have to change anything if the reliabilty increased above
        # the $reliable value and when it drops below.
        # ########

        if( ( ($rely< $reliable)&&($rely{$from}>=$reliable) ) || 
            ( ($rely>=$reliable)&&($rely{$from}< $reliable) ) )
        {
            

            print STDERR "STATUS OF HARDROUTE:$hostname{$from} changed\n" if $debug;
            if( ($rely>=$reliable) )
            {
                # ########
                # The server is now reliable: Let's ask him what connections
                # he can offer us.
                # ########

                if( $router{$from} )
                {
                    $message=&encodehosts($from);
                    &rmessage($router{$from},$myhostaddr,++$mymessageid,"delete broadcast",$message);
                }
                $router{$from}=$from;
                $rely{$from}=$rely;
                $speed{$from}=$speed;
                $hops{$from}=0;
                &rmessage($from,$myhostaddr,++$myrmessageid,"whom do you know",undef);
                foreach $mom (keys(%hardrouted))
                {
                    if( ($mom ne $from) && ($rely{$mom}>=$reliable) )
                    {
                        $message=pack("a4 s f",$from,0,$speed);
                        &rmessage($mom,$myhostaddr,++$myrmessageid,"i know",$message);
                    }
                }
                
                # ########
                # Let's get our Messagehandler informed, that a server got reliable
                # ########
                &HANDLER_ConnectionEstablished($from);

            }
            else
            {
                
                # ########
                # in case the server is no longer reliable we have to remove
                # the routings and broadcastings. Furthermore we have to ask
                # the other remaining reliable servers which routes they know,
                # so that we are able to regain almost every virtual connection
                # to existing servers in the net.
                # ########

                $rely{$from}=$rely;
                &removehost($from);
                $removerouting{$from}=1;
                foreach $mom (keys(%hardrouted))
                {
                    if( ($mom ne $from) && ($rely{$mom}>=$reliable) )
                    {
                        $message=pack("a4 s f",$from,0,$speed);
                        &rmessage($mom,$myhostaddr,++$myrmessageid,"whom do you know",undef);
                    }
                }
                
            }
        }
    }
    &printrouting if $debug;
}


