# ########
# This file includes all the routines we have to communicate with
# servers on a higher level 
#
# global vars should begin with a H_ to make them distinguishable 
#
#   As described in the API there are these fine routines , which get
#   called by the server to provide an easier way to program a distributed
#   system
#
#   &HANDLER_Init
#   &HANDLER_Heartbeat
#   &HANDLER_ConnectionEstablished($from)
#   &HANDLER_GotCommand($from,$origin,$origid,$command,$message)
#   &HANDLER_GotBroadcast($from,$origin,$origid,$message)
#
#
# ########
$isnotused=1;#used for proper includeing 


$startdir="..";

sub HANDLER_Init
{
    local(@files,$mom);

    $H_myhostaddr=join(".",unpack("C C C C",$myhostaddr));
    die "Sorry directory $startdir/$myhostname not found\n" if !-d "$startdir/$myhostname";

    if( (-s "$startdir/$myhostname/removed") && (open(H_REMOVED,"$startdir/$myhostname/removed")) )
    {
        while(<H_REMOVED>)
        {
            s/\n//;
            $H_removed{$_}=1;
            unlink("$startdir/$myhostname/$_") if -f"$startdir/$myhostname/$_";
        }
        close(H_REMOVED);   
    }
    
    opendir(H_DIR,"$startdir/$myhostname/messages") || die "can't open directory $startdir/$myhostname/messages\n";
    @files=grep(/MSG_\d+\-\d+\.\d+\.\d+\.\d+/,readdir(H_DIR));
    closedir(H_DIR);
    foreach (@files)
    {
        if(/MSG_(\d+)\-(\d+)\.(\d+)\.(\d+)\.(\d+)/)
        {
            $mom=pack("I C C C C",$1,$2,$3,$4,$5);
            $H_message{$mom}=1;
        }
    }

    return undef;
}


sub HANDLER_Heartbeat
{
    local($sec,$min,$hour,$mday,$mon,$year);
    local($header,$footer,$message);
    local($time,$host,@array);
    undef $message;
    undef $command;

    if( open(H_NEWMESSAGE,"$startdir/$myhostname/message"))
    {
        $time=time;
        $header=pack("I a4",$time,$myhostaddr);
        while(<H_NEWMESSAGE>)
        {
            $footer.=$_;
        }
        close(H_NEWMESSAGE);
        system "mv $startdir/$myhostname/message $startdir/$myhostname/messages/MSG_$time-$H_myhostaddr";    
        $message=$header.$footer;
        $command="broadcast";
    }

    if( open(H_NEWREMOVE,"$startdir/$myhostname/remove") )
    {
        while(<H_NEWREMOVE>)    
        {
            s/\n//;
            push(@array,$_);
            $H_removed{$_}=1;
            unlink("$startdir/$myhostname/messages/$_");
        }
        close(H_NEWREMOVE);
        unlink("$startdir/$myhostname/remove");
        open(H_REMOVED,"> $startdir/$myhostname/removed") || die "error opening removed file\n";
        print H_REMOVED join("\n",keys(%H_removed));
        close(H_REMOVED);
        $message=join("\n",@array);
        $command="command:these are removed";
    }

    if($command)
    {
        foreach $host ( keys(%hardrouted) )
        {
            if($rely{$host}>=$reliable)
            {
                $mymessageid++;
                &rmessage($host,$myhostaddr,$mymessageid,$command,$message);
            }
        }
    }

    return undef;
}





sub HANDLER_ConnectionEstablished
{
    local($from)=@_;
    &sendCommand($from,"send removed",undef);
    &sendCommand($from,"send stored",undef);
    return undef;
}



sub HANDLER_GotCommand
{
    local($from,$origin,$origid,$command,$message)=@_;

    local($mom,$reply,$time,$addr);
    $_=$command;

    if(/^send removed$/)
    {
        $message=join("\n",keys(%H_removed));
        &sendCommand($from,"these are removed",$message) if $message;
        return undef;
    }

    if(/^these are removed$/)
    {
        @array=split(/\n/,$message);
        return undef if ! @array;

        foreach $mom ( @array )
        {
            $H_removed{$mom}=1;
            $H_message{$mom}=0;
            unlink("$startdir/$myhostname/messages/$mom") if -f "$startdir/$myhostname/messages/$mom";
        }
        open(H_REMOVED,"> $startdir/$myhostname/removed") || die "error opening removed file\n";
        print H_REMOVED join("\n",keys(%H_removed));
        close(H_REMOVED);
        return undef;
    }
    

    if(/^send stored$/)
    {
        undef $reply;

        foreach $mom (keys(%H_message))
        {
            $reply.=$mom if $H_message{$mom};
        }
        &sendCommand($from,"i know these",$reply) if $reply;
        return undef;
    }

    if(/^i know these$/)
    {
        undef $reply;

        while($message)
        {
            $mom=substr($message,0,8);
            $message=substr($message,8);
            if( (!$H_removed{$mom}) && (!$H_message{$mom}) )
            {
                $reply.=$mom;
            }
        }
        &sendCommand($from,"request",$reply) if $reply;
        return undef;
    }


    if(/^request$/)
    {
        while($message)
        {
            undef $reply;
            $mom=substr($message,0,8);
            $message=substr($message,8);
            ($time,$addr)=unpack("I a4",$mom);
            $name="MSG_$time-".join(".",unpack("C C C C",$addr));
            if(open(H_MESSAGE,"$startdir/$myhostname/messages/$name"))
            {
                $reply=$mom;
                while(<H_MESSAGE>)
                {
                    $reply.=$_;
                }
                close(H_MESSAGE);
            }
            &sendCommand($from,"requested message",$reply) if $reply;
        }
        return undef;
    }

    if(/^requested message$/)
    {
        $header=substr($message,0,8);
        $message=substr($message,8);
        ($time,@addr)=unpack("I C4",$header);
        $name="MSG_$time-".join(".",@addr);
    
        open(NEWMESSAGE,"> $startdir/$myhostname/messages/$name") || die "can't open $startdir/$myhostname/messages/$name\n";
        print NEWMESSAGE $message;
        close(NEWMESSAGE);
    
        return undef;
    }

    return undef;
}



sub HANDLER_GotBroadcast
{
    local($from,$origin,$origid,$message)=@_;   
    local($header,$time,$addr,$name);
    $header=substr($message,0,8);
    $message=substr($message,8);

    ($time,@addr)=unpack("I C4",$header);
    $name="MSG_$time-".join(".",@addr);

    open(NEWMESSAGE,"> $startdir/$myhostname/messages/$name") || die "can't open $startdir/$myhostname/messages/$name\n";
    print NEWMESSAGE $message;
    close(NEWMESSAGE);

    return undef;
}


